import os
from pathlib import Path
from fpcmci.basics.constants import *


def cls():
    """
    Clear terminal
    """
    os.system('cls' if os.name == 'nt' else 'clear')


def get_selectorpath(resfolder):
    """
    Return log file path

    Args:
        resfolder (str): result folder

    Returns:
        (str, str): log file path, dependency image file path
    """
    Path(SEP.join([RESULTS_FILENAME, resfolder])).mkdir(parents=True, exist_ok=True)
    return SEP.join([RESULTS_FILENAME, resfolder, LOG_FILENAME]), SEP.join([RESULTS_FILENAME, resfolder, DEP_FILENAME])


def create_results_folder():
    """
    Creates results folder if doesn't exist
    """
    Path(RESULTS_FILENAME).mkdir(parents=True, exist_ok=True)


def get_validatorpaths(resfolder):
    """
    Creates resfolder if doesn't exist

    Args:
        resfolder (str): result folder name

    Returns:
        (str, str, str): result.pkl file path, dag file path, ts_dag file path
    """
    Path(SEP.join([RESULTS_FILENAME, resfolder])).mkdir(parents=True, exist_ok=True)
    return SEP.join([RESULTS_FILENAME, resfolder, RES_FILENAME]), SEP.join([RESULTS_FILENAME, resfolder, DAG_FILENAME]), SEP.join([RESULTS_FILENAME, resfolder, TSDAG_FILENAME])


from fpcmci.selection_methods.SelectionMethod import SelectionMethod, CTest
from fpcmci.CPrinter import CP
from scipy import stats, linalg
import numpy as np

class ParCorr(SelectionMethod):
    """
    Feature selection method based on Partial Correlation analysis
    """
    def __init__(self):
        """
        ParCorr class contructor
        """
        super().__init__(CTest.Corr)


    def get_residual(self, covar, target):
        """
        Calculate residual of the target variable obtaining conditioning on the covar variables

        Args:
            covar (np.array): conditioning variables
            target (np.array): target variable

        Returns:
            (np.array): residual
        """
        beta = np.linalg.lstsq(covar, target, rcond=None)[0]
        return target - np.dot(covar, beta)


    def partial_corr(self, X, Y, Z):
        """
        Calculate Partial correlation between X and Y conditioning on Z

        Args:
            X (np.array): source candidate variable
            Y (np.array): target variable
            Z (np.array): conditioning variable

        Returns:
            (float, float): partial correlation, p-value
        """

        pcorr, pval = stats.pearsonr(self.get_residual(Z, X), self.get_residual(Z, Y))

        return pcorr, pval

    def compute_dependencies(self):
        """
        compute list of dependencies for each target by partial correlation analysis

        Returns:
            (dict): dictonary(TARGET: list SOURCES)
        """
        CP.info("\n##")
        CP.info("## " + self.name + " analysis")
        CP.info("##")

        for lag in range(self.min_lag, self.max_lag + 1):
            for target in self.data.features:
                CP.info("\n## Target variable: " + target)
                candidates = self.data.features

                Y = np.array(self.data.d[target][lag:])

                while candidates:
                    tmp_res = None
                    covars = self._get_sources(target)
                    Z = np.array(self.data.d[covars][:-lag])

                    for candidate in candidates:
                        X = np.array(self.data.d[candidate][:-lag])
                        score, pval = self.partial_corr(X, Y, Z)
                        if pval < self.alpha and (tmp_res is None or abs(tmp_res[1]) < abs(score)):
                            tmp_res = (candidate, score, pval)

                    if tmp_res is not None: 
                        self._add_dependecies(target, tmp_res[0], tmp_res[1], tmp_res[2], lag)
                        candidates.remove(tmp_res[0])
                    else:
                        break
        return self.result


from abc import ABC, abstractmethod
from enum import Enum
import pandas as pd


class SSMode(Enum):
    WSDynamic = 'Dynamic-size moving window'
    WSStatic = 'Static-size moving window'
    WSFFTStatic = 'FFT static-size moving window'
    Static = 'Static'
    

class SubsamplingMethod(ABC):
    """
    SubsamplingMethod abstract class
    """
    def __init__(self, ssmode: SSMode):
        self.ssmode = ssmode
        self.df = None

    
    def initialise(self, dataframe: pd.DataFrame):
        """
        Initialise class by setting the dataframe to subsample

        Args:
            dataframe (pd.DataFrame): _description_
        """
        self.df = dataframe


    @abstractmethod
    def run(self):
        """
        Run subsampler
        """
        pass



from abc import ABC, abstractmethod
from copy import deepcopy
from fpcmci.preprocessing.subsampling_methods.moving_window import MovingWindow

    
class EntropyBasedMethod(ABC):
    """
    EntropyBasedMethod abstract class
    """
    def __init__(self, threshold):
        self.windows = list()
        self.segments = list()
        self.threshold = threshold


    def create_rounded_copy(self):
        """
        Create deepcopy of the dataframe but with rounded values

        Returns:
            (pd.DataFrame): rounded dataframe
        """
        de = deepcopy(self.df)
        de = de.round(1)
        return de


    def __normalization(self):
        """
        Normalize entropy for each moving window
        """
        max_e = max([mw.entropy for mw in self.windows])
        for mw in self.windows:
            mw.entropy = mw.entropy / max_e


    def moving_window_analysis(self):
        """
        Compute dataframe entropy on moving windows
        """
        de = self.create_rounded_copy()

        for ll, rl in self.segments:
            # Create moving window
            mw_df = de.values[ll: rl]

            # Build a Moving Window
            mw = MovingWindow(mw_df)

            # Compute entropy
            mw.get_entropy()

            # Compute optimal number of samples
            mw.optimal_sampling(self.threshold)

            # Collect result in a list
            self.windows.append(mw)

        # Entropy normalization
        self.__normalization()


    # def extract_data(self):
    #     """
    #     Extract plottable data from moving window analysis
    #     """
    #     # Entropies and samples numbers list
    #     self.__entropy_list = [mw.entropy for mw in self.__window_list]
    #     self.__sample_number_list = [mw.opt_size for mw in self.__window_list]
    #     self.__original_size = [mw.T for mw in self.__window_list]
    #     self.num_samples = sum(self.__sample_number_list)

    #     # Make entropy and sample array plottable
    #     self.__pretty_signals()


    # def __pretty_signals(self):
    #     """
    #     Make entropy list and sample number list plottable
    #     """
    #     _pretty_entropy = []
    #     _pretty_sample_number = []
    #     _pretty_original_size = []
    #     for i, mw in enumerate(self.__window_list):
    #         _pretty_entropy += np.repeat(self.__entropy_list[i], mw.T).tolist()
    #         _pretty_sample_number += np.repeat(self.__sample_number_list[i], mw.T).tolist()
    #         _pretty_original_size += np.repeat(self.__original_size[i], mw.T).tolist()
    #     self.__entropy_list = _pretty_entropy
    #     self.__sample_number_list = _pretty_sample_number
    #     self.__original_size = _pretty_original_size

    #     _diff = self.df.shape[0] - len(self.__entropy_list)
    #     if _diff != 0:
    #         self.__entropy_list = np.append(self.__entropy_list, [self.__entropy_list[-1]] * _diff)
    #         self.__sample_number_list = np.append(self.__sample_number_list, [self.__sample_number_list[-1]] * _diff)


    def extract_indexes(self):
        """
        Extract a list of indexes corresponding to the samples
        selected by the subsampling procedure
        """
        _sample_index_list = list()
        for i, mw in enumerate(self.windows):
            sum_ws = sum([wind.T for wind in self.windows[:i]])
            sample_index = [si + sum_ws for si in mw.opt_samples_index]
            _sample_index_list += sample_index
        return _sample_index_list


    @abstractmethod
    def dataset_segmentation(self):
        """
        abstract method
        """
        pass


from .EntropyBasedMethod import EntropyBasedMethod
from .moving_window import MovingWindow
from .Static import Static
from .SubsamplingMethod import SubsamplingMethod
from .WSDynamic import WSDynamic
from .WSFFTStatic import WSFFTStatic
from .WSStatic import WSStatic


from enum import Enum
from fpcmci.selection_methods.SelectionMethod import SelectionMethod, CTest, _suppress_stdout
from idtxl.multivariate_mi import MultivariateMI
from idtxl.data import Data

class MIestimator(Enum):
    Gaussian = 'JidtGaussianCMI'
    Kraskov = 'JidtKraskovCMI'


class MI(SelectionMethod):
    """
    Feature selection method based on Mutual Information analysis
    """
    def __init__(self, estimator: MIestimator):
        """
        MI class contructor

        Args:
            estimator (MIestimator): Gaussian/Kraskov
        """
        super().__init__(CTest.MI)
        self.estimator = estimator

    def compute_dependencies(self):
        """
        compute list of dependencies for each target by mutual information analysis

        Returns:
            (dict): dictonary(TARGET: list SOURCES)
        """
        with _suppress_stdout():
            data = Data(self.d.values, dim_order='sp') # sp = samples(row) x processes(col)

            network_analysis = MultivariateMI()
            settings = {'cmi_estimator': self.estimator.value,
                        'max_lag_sources': self.max_lag,
                        'min_lag_sources': self.min_lag,
                        'alpha_max_stats': self.alpha,
                        'alpha_min_stats': self.alpha,
                        'alpha_omnibus': self.alpha,
                        'alpha_max_seq': self.alpha,
                        'verbose': False}
            results = network_analysis.analyse_network(settings=settings, data=data)

        for t in results._single_target.keys():
            sel_sources = [s[0] for s in results._single_target[t]['selected_vars_sources']]
            if sel_sources:
                sel_sources_lag = [s[1] for s in results._single_target[t]['selected_vars_sources']]
                sel_sources_score = results._single_target[t]['selected_sources_mi']
                sel_sources_pval = results._single_target[t]['selected_sources_pval']
                for s, score, pval, lag in zip(sel_sources, sel_sources_score, sel_sources_pval, sel_sources_lag):
                    self._add_dependecies(self.features[t], self.features[s], score, pval, lag)

        return self.result


from matplotlib import pyplot as plt
import numpy as np
import networkx as nx
from netgraph import Graph
from fpcmci.basics.constants import *


def dag(res,
        node_layout = 'dot',
        min_width = 1,
        max_width = 5,
        min_score = 0,
        max_score = 1,
        node_size = 8,
        node_color = 'orange',
        edge_color = 'grey',
        font_size = 12,
        label_type = LabelType.Lag,
        save_name = None):
    """
    build a dag

    Args:
        res (dict): dependencies result
        node_layout (str, optional): Node layout. Defaults to 'dot'.
        min_width (int, optional): minimum linewidth. Defaults to 1.
        max_width (int, optional): maximum linewidth. Defaults to 5.
        min_score (int, optional): minimum score range. Defaults to 0.
        max_score (int, optional): maximum score range. Defaults to 1.
        node_size (int, optional): node size. Defaults to 8.
        node_color (str, optional): node color. Defaults to 'orange'.
        edge_color (str, optional): edge color. Defaults to 'grey'.
        font_size (int, optional): font size. Defaults to 12.
        label_type (LabelType, optional): enum to set whether to show the lag time (LabelType.Lag) or the strength (LabelType.Score) of the dependencies on each link/node or not showing the labels (LabelType.NoLabels). Default LabelType.Lag.
        save_name (str, optional): Filename path. If None, plot is shown and not saved. Defaults to None.
    """

    G = nx.DiGraph()

    # NODES DEFINITION
    G.add_nodes_from(res.keys())
    
    # BORDER LINE
    border = dict()
    for t in res.keys():
        border[t] = 0
        for s in res[t]:
            if t == s[SOURCE]:
                border[t] = max(__scale(s[SCORE], min_width, max_width, min_score, max_score), border[t])
    
    # BORDER LABEL
    node_label = None
    if label_type == LabelType.Lag or label_type == LabelType.Score:
        node_label = {t: [] for t in res.keys()}
        for t in res.keys():
            for s in res[t]:
                if t == s[SOURCE]:
                    if label_type == LabelType.Lag:
                        node_label[t].append(s[LAG])
                    elif label_type == LabelType.Score:
                        node_label[t].append(round(s[SCORE], 3))
            node_label[t] = ",".join(str(s) for s in node_label[t])


    # EDGE DEFINITION
    edges = [(s[SOURCE], t) for t in res.keys() for s in res[t] if t != s[SOURCE]]
    G.add_edges_from(edges)
    
    # EDGE LINE
    edge_width = {(s[SOURCE], t): 0 for t in res.keys() for s in res[t] if t != s[SOURCE]}
    for t in res.keys():
        for s in res[t]:
            if t != s[SOURCE]:
                edge_width[(s[SOURCE], t)] = max(__scale(s[SCORE], min_width, max_width, min_score, max_score), edge_width[(s[SOURCE], t)])
    
    # EDGE LABEL
    edge_label = None
    if label_type == LabelType.Lag or label_type == LabelType.Score:
        edge_label = {(s[SOURCE], t): [] for t in res.keys() for s in res[t] if t != s[SOURCE]}
        for t in res.keys():
            for s in res[t]:
                if t != s[SOURCE]:
                    if label_type == LabelType.Lag:
                        edge_label[(s[SOURCE], t)].append(s[LAG])
                    elif label_type == LabelType.Score:
                        edge_label[(s[SOURCE], t)].append(round(s[SCORE], 3))
        for k in edge_label.keys():
            edge_label[k] = ",".join(str(s) for s in edge_label[k])

    fig, ax = plt.subplots(figsize=(8,6))

    if edges:
        a = Graph(G, 
                node_layout = node_layout,
                node_size = node_size,
                node_color = node_color,
                node_labels = node_label,
                node_edge_width = border,
                node_label_fontdict = dict(size=font_size),
                node_edge_color = edge_color,
                node_label_offset = 0.1,
                node_alpha = 1,
                
                arrows = True,
                edge_layout = 'curved',
                edge_label = label_type != LabelType.NoLabels,
                edge_labels = edge_label,
                edge_label_fontdict = dict(size=font_size),
                edge_color = edge_color, 
                edge_width = edge_width,
                edge_alpha = 1,
                edge_zorder = 1,
                edge_label_position = 0.35,
                edge_layout_kwargs = dict(bundle_parallel_edges = False, k = 0.05))
        
        nx.draw_networkx_labels(G, 
                                pos = a.node_positions,
                                labels = {n: n for n in G},
                                font_size = font_size)

    if save_name is not None:
        plt.savefig(save_name, dpi = 300)
    else:
        plt.show()
               
        
def ts_dag(res,
           tau,
           min_width = 1,
           max_width = 5,
           min_score = 0,
           max_score = 1,
           node_size = 8,
           node_color = 'orange',
           edge_color = 'grey',
           font_size = 12,
           save_name = None):
    """
    build a timeseries dag

    Args:
        res (dict): dependencies result
        tau (int): max time lag
        min_width (int, optional): minimum linewidth. Defaults to 1.
        max_width (int, optional): maximum linewidth. Defaults to 5.
        min_score (int, optional): minimum score range. Defaults to 0.
        max_score (int, optional): maximum score range. Defaults to 1.
        node_size (int, optional): node size. Defaults to 8.
        node_color (str, optional): node color. Defaults to 'orange'.
        edge_color (str, optional): edge color. Defaults to 'grey'.
        font_size (int, optional): font size. Defaults to 12.
        save_name (str, optional): Filename path. If None, plot is shown and not saved. Defaults to None.
    """
    
    # add nodes
    G = nx.grid_2d_graph(tau + 1, len(res.keys()))
    pos = dict()
    for n in G.nodes():
        if n[0] == 0:
            pos[n] = (n[0], n[1]/2)
        else:
            pos[n] = (n[0] + .5, n[1]/2)
    scale = max(pos.values())
    G.remove_edges_from(G.edges())

    # edges definition
    edges = list()
    edge_width = dict()
    for t in res.keys():
        for s in res[t]:
            s_index = len(res.keys())-1 - list(res.keys()).index(s[SOURCE])
            t_index = len(res.keys())-1 - list(res.keys()).index(t)
            s_node = (tau - s[LAG], s_index)
            t_node = (tau, t_index)
            edges.append((s_node, t_node))
            edge_width[(s_node, t_node)] = __scale(s[SCORE], min_width, max_width, min_score, max_score)
    G.add_edges_from(edges)

    # label definition
    labeldict = {}
    for n in G.nodes():
        if n[0] == 0:
            labeldict[n] = list(res.keys())[len(res.keys()) - 1 - n[1]]

    fig, ax = plt.subplots(figsize=(8,6))

    # time line text drawing
    pos_tau = set([pos[p][0] for p in pos])
    max_y = max([pos[p][1] for p in pos])
    for p in pos_tau:
        if abs(int(p) - tau) == 0:
            ax.text(p, max_y + .3, r"$t$", horizontalalignment='center', fontsize=font_size)
        else:
            ax.text(p, max_y + .3, r"$t-" + str(abs(int(p) - tau)) + "$", horizontalalignment='center', fontsize=font_size)

    Graph(G,
          node_layout = {p : np.array(pos[p]) for p in pos},
          node_size = node_size,
          node_color = node_color,
          node_labels = labeldict,
          node_label_offset = 0,
          node_edge_width = 0,
          node_label_fontdict = dict(size=font_size),
          node_alpha = 1,
          
          arrows = True,
          edge_layout = 'curved',
          edge_label = False,
          edge_color = edge_color, 
          edge_width = edge_width,
          edge_alpha = 1,
          edge_zorder = 1,
          scale = (scale[0] + 2, scale[1] + 2))

    if save_name is not None:
        plt.savefig(save_name, dpi = 300)
    else:
        plt.show()


def __scale(score, min_width, max_width, min_score = 0, max_score = 1):
    """
    Scales the score of the cause-effect relationship strength to a linewitdth

    Args:
        score (float): score to scale
        min_width (float): minimum linewidth
        max_width (float): maximum linewidth
        min_score (int, optional): minimum score range. Defaults to 0.
        max_score (int, optional): maximum score range. Defaults to 1.

    Returns:
        (float): scaled score
    """
    return ((score - min_score) / (max_score - min_score)) * (max_width - min_width) + min_width



from fpcmci.selection_methods.SelectionMethod import SelectionMethod, CTest
from sklearn.feature_selection import f_regression
from fpcmci.CPrinter import CP

class Corr(SelectionMethod):
    """
    Feature selection method based on Correlation analysis
    """
    def __init__(self):
        """
        Corr contructor class
        """
        super().__init__(CTest.Corr)


    def compute_dependencies(self):
        """
        compute list of dependencies for each target by correlation analysis

        Returns:
            (dict): dictonary(TARGET: list SOURCES)
        """
        CP.info("\n##")
        CP.info("## " + self.name + " analysis")
        CP.info("##")

        for lag in range(self.min_lag, self.max_lag + 1):
            for target in self.data.features:
                CP.info("\n## Target variable: " + target)

                X, Y = self._prepare_ts(target, lag)
                scores, pval = f_regression(X, Y)
                
                # Filter on pvalue
                f = pval < self.alpha

                # Result of the selection
                sel_sources, sel_sources_score, sel_sources_pval = X.columns[f].tolist(), scores[f].tolist(), pval[f].tolist()

                for s, score, pval in zip(sel_sources, sel_sources_score, sel_sources_pval):
                    self._add_dependecies(target, s, score, pval, lag)

        return self.result


from abc import ABC, abstractmethod
from enum import Enum
from contextlib import contextmanager
import sys, os
from fpcmci.preprocessing.data import Data
from fpcmci.basics.constants import *


class CTest(Enum):
    Corr = "Correlation"
    MI = "Mutual Information"
    TE = "Transfer Entropy"


@contextmanager
def _suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:  
            yield
        finally:
            sys.stdout = old_stdout


class SelectionMethod(ABC):
    """
    SelectionMethod abstract class
    """
    def __init__(self, ctest):
        self.ctest = ctest
        self.data = None
        self.alpha = None
        self.min_lag = None
        self.max_lag = None
        self.result = dict()


    @property
    def name(self):
        """
        Returns Selection Method name

        Returns:
            (str): Selection Method name
        """
        return self.ctest.value


    def initialise(self, data: Data, alpha, min_lag, max_lag):
        """
        Initialises the selection method

        Args:
            data (Data): Data
            alpha (float): significance threshold
            min_lag (int): min lag time
            max_lag (int): max lag time
        """
        self.data = data
        self.alpha = alpha
        self.min_lag = min_lag
        self.max_lag = max_lag
        self.result = {f:list() for f in self.data.features}


    @abstractmethod
    def compute_dependencies(self) -> dict:
        """
        abstract method
        """
        pass


    def _prepare_ts(self, target, lag, apply_lag = True, consider_autodep = True):
        """
        prepare the dataframe to the analysis

        Args:
            target (str): name target var
            lag (int): lag time to apply
            apply_lag (bool, optional): True if you want to apply the lag, False otherwise. Defaults to True.

        Returns:
            tuple(DataFrame, DataFrame): source and target dataframe
        """
        if not consider_autodep:
            if apply_lag:
                Y = self.data.d[target][lag:]
                X = self.data.d.loc[:, self.data.d.columns != target][:-lag]
            else:
                Y = self.data.d[target]
                X = self.data.d.loc[:, self.data.d.columns != target]
        else:
            if apply_lag:
                Y = self.data.d[target][lag:]
                X = self.data.d[:-lag]
            else:
                Y = self.data.d[target]
                X = self.data.d
        return X, Y


    def _get_sources(self, t):
        """
        Return target sources

        Args:
            t (str): target variable name

        Returns:
            list(str): list of target sources
        """
        return [s[SOURCE] for s in self.result[t]]


    def _add_dependecies(self, t, s, score, pval, lag):
        """
        Adds found dependency from source (s) to target (t) specifying the 
        score, pval and the lag

        Args:
            t (str): target feature name
            s (str): source feature name
            score (float): selection method score
            pval (float): pval associated to the dependency
            lag (int): lag time of the dependency
        """
        self.result[t].append({SOURCE:s, 
                               SCORE:score,
                               PVAL:pval, 
                               LAG:lag})
        str_s = "(" + s + " -" + str(lag) + ")"
        str_arrow = " --> "
        str_t = "(" + t + ")"
        str_score = "|score: " + "{:.3f}".format(score)
        str_pval = "|pval: " + "{:.3f}".format(pval)
        print('{:<20s}{:<10s}{:<10s}{:<20s}{:<20s}'.format(str_s, str_arrow, str_t, str_score, str_pval))



from enum import Enum

class CPLevel(Enum):
    NONE = 0
    WARN = 1
    INFO = 2
    DEBUG = 3


class CPrinter():
    def __init__(self):
        self.verbosity = None
    
    def set_verbosity(self, verbosity: CPLevel):
        """
        set verbosity level

        Args:
            verbosity (CPLevel): verbosity level
        """
        self.verbosity = verbosity
        
    def warning(self, msg: str):
        """
        write message iff verbosity >= WARN

        Args:
            msg (str): massage to write
        """
        if self.verbosity.value >= CPLevel.WARN.value: print(msg)

    def info(self, msg: str):
        """
        write message iff verbosity >= INFO

        Args:
            msg (str): massage to write
        """
        if self.verbosity.value >= CPLevel.INFO.value: print(msg)

    def debug(self, msg: str):
        """
        write message iff verbosity >= DEBUG

        Args:
            msg (str): massage to write
        """
        if self.verbosity.value >= CPLevel.DEBUG.value: print(msg)

CP = CPrinter()


import copy
import pickle
import matplotlib.pyplot as plt
import numpy as np
from tigramite.independence_tests.independence_tests_base import CondIndTest
import sys
from fpcmci.selection_methods.SelectionMethod import SelectionMethod
from fpcmci.CPrinter import CPLevel, CP
from fpcmci.basics.constants import *
from fpcmci.basics.logger import Logger
import fpcmci.basics.utils as utils
from fpcmci.PCMCI import PCMCI
from fpcmci.preprocessing.data import Data 


class FPCMCI():
    """
    FPCMCI class.

    FPCMCI is a causal feature selector framework for large-scale time series
    datasets. Sarting from a Data object and it selects the main features
    responsible for the evolution of the analysed system. Based on the selected features,
    the framework outputs a causal model.
    """

    def __init__(self, 
                 data: Data, 
                 min_lag, max_lag, 
                 sel_method: SelectionMethod, val_condtest: CondIndTest, 
                 verbosity: CPLevel, 
                 f_alpha = 0.05, 
                 pcmci_alpha = 0.05, 
                 resfolder = None,
                 neglect_only_autodep = False):
        """
        FPCMCI class contructor

        Args:
            data (Data): data to analyse
            min_lag (int): minimum time lag
            max_lag (int): maximum time lag
            sel_method (SelectionMethod): selection method
            val_condtest (CondIndTest): validation method
            verbosity (CPLevel): verbosity level
            f_alpha (float, optional): filter significance level. Defaults to 0.05.
            pcmci_alpha (float, optional): PCMCI significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
        """
        
        self.data = data
        self.f_alpha = f_alpha
        self.pcmci_alpha = pcmci_alpha
        self.min_lag = min_lag
        self.max_lag = max_lag
        self.sel_method = sel_method
        self.filter_dependencies = None
        self.o_filter_dependencies = None
        self.causal_model = None
        self.result = None
        self.neglect_only_autodep = neglect_only_autodep

        self.dependency_path = None
        if resfolder is not None:
            utils.create_results_folder()
            logpath, self.dependency_path = utils.get_selectorpath(resfolder)
            sys.stdout = Logger(logpath)
        
        self.validator = PCMCI(data, self.pcmci_alpha, min_lag, max_lag, val_condtest, resfolder, verbosity)       
        CP.set_verbosity(verbosity)


    def run_filter(self):
        """
        Run filter method
        """
        CP.info("\n")
        CP.info(DASH)
        CP.info("Selecting relevant features among: " + str(self.data.features))
        CP.info("Selection method: " + self.sel_method.name)
        CP.info("Significance level: " + str(self.f_alpha))
        CP.info("Max lag time: " + str(self.max_lag))
        CP.info("Min lag time: " + str(self.min_lag))
        CP.info("Data length: " + str(self.data.T))

        self.sel_method.initialise(self.data, self.f_alpha, self.min_lag, self.max_lag)
        self.filter_dependencies = self.sel_method.compute_dependencies()
        self.o_filter_dependecies = copy.deepcopy(self.filter_dependencies)


    def run_pcmci(self):
        """
        Run PCMCI
        
        Returns:
            list(str): list of selected variable names
            dict(str:list(tuple)): causal model
        """
        CP.info("Significance level: " + str(self.pcmci_alpha))
        CP.info("Max lag time: " + str(self.max_lag))
        CP.info("Min lag time: " + str(self.min_lag))
        CP.info("Data length: " + str(self.data.T))

        # causal model
        self.validator.data = self.data
        self.validator.run()
        self.causal_model = self.validator.dependencies
        
        self.result = self.data.features
        
        self.save_validator_res()
        
        return self.result, self.causal_model

    
    def run(self):
        """
        Run Selector and Validator
        
        Returns:
            list(str): list of selected variable names
            dict(str:list(tuple)): causal model
        """
        
        self.run_filter()        
            
        # list of selected features based on dependencies
        tmp_sel_features = self.get_selected_features()
        if not tmp_sel_features:
            return self.result

        # shrink dataframe d and dependencies by the selector result
        self.shrink(tmp_sel_features)
        
        # selected links to check by the validator
        link_assumptions = self.__get_link_assumptions()
            
        # causal model on selected links
        self.validator.data = self.data
        pcmci_result = self.validator.run(link_assumptions)
        
        # application of the validator result to the filter_dependencies field
        self.__apply_validator_result(pcmci_result)
        
        self.result = self.get_selected_features()
        # shrink dataframe d and dependencies by the validator result
        self.shrink(self.result)
        
        # final causal model
        self.causal_model = self.validator.dependencies
        self.save_validator_res()
        
        CP.info("\nFeature selected: " + str(self.result))
        return self.result, self.causal_model
    
    
    def get_causal_matrix(self):
        """
        Returns a dictionary with keys the lags and values the causal matrix containing the causal weights between targets (rows) and sources (columns)

        Returns:
            dict/np.ndarray: causal matrix per 
        """
        cm_per_lag = {lag : np.zeros((len(self.result), len(self.result))) for lag in range(self.min_lag, self.max_lag + 1)}
        vars = ['$' + var +'$' for var in self.result]
        for lag in cm_per_lag:
            for var in vars:
                for source in self.causal_model[var]:
                    if source[LAG] == lag: cm_per_lag[lag][vars.index(var)][vars.index(source[SOURCE])] = source[SCORE]
        if len(cm_per_lag) == 1: return list(cm_per_lag.values())[0]
        return cm_per_lag
    
    
    def get_SCM(self):
        """
        Return Structural Causal Model

        Raises:
            ValueError: "Causal Model not estimated yet" if self.causal_model is None

        Returns:
            dict(str:list(tuple)): SCM of the causal model in the format "target": [(source, lag) ...] (e.g. "$X0$" : [("$X0$", -1), ("$X1$", -2)]) 
        """
        if self.causal_model is None:
            raise ValueError("Causal Model not estimated yet.")
        scm = {v: list() for v in self.data.pretty_features}
        for t in self.causal_model.keys():
            for s in self.causal_model[t]:
                scm[t].append((s[SOURCE], -s[LAG])) 
        return scm
    

    def shrink(self, sel_features):
        """
        Wrapper in order to shrink data.d and dependencies

        Args:
            sel_features (list(str)): list of selected features
        """
        self.data.shrink(sel_features)
        self.__shrink_dependencies()
    
    
    def save_validator_res(self):
        """
        Saves dag plot if resfolder has been set otherwise it shows the figure
        """
        if self.result:
            self.validator.save_result()
        else:
            CP.warning("Result impossible to save: no feature selected")
    
    
    def dag(self,
            node_layout = 'dot',
            min_width = 1,
            max_width = 5,
            min_score = 0,
            max_score = 1,
            node_size = 8,
            node_color = 'orange',
            edge_color = 'grey',
            font_size = 12,
            label_type = LabelType.Lag):
        """
        Saves dag plot if resfolder has been set otherwise it shows the figure
        
        Args:
            node_layout (str, optional): Node layout. Defaults to 'dot'.
            min_width (int, optional): minimum linewidth. Defaults to 1.
            max_width (int, optional): maximum linewidth. Defaults to 5.
            min_score (int, optional): minimum score range. Defaults to 0.
            max_score (int, optional): maximum score range. Defaults to 1.
            node_size (int, optional): node size. Defaults to 8.
            node_color (str, optional): node color. Defaults to 'orange'.
            edge_color (str, optional): edge color. Defaults to 'grey'.
            font_size (int, optional): font size. Defaults to 12.
            label_type (LabelType, optional): enum to set whether to show the lag time (LabelType.Lag) or the strength (LabelType.Score) of the dependencies on each link/node or not showing the labels (LabelType.NoLabels). Default LabelType.Lag.
        """
        
        if self.result:
            self.validator.build_dag(node_layout,
                                     min_width, 
                                     max_width,
                                     min_score,
                                     max_score,
                                     node_size,
                                     node_color,
                                     edge_color,
                                     font_size,
                                     label_type)
        else:
            CP.warning("Dag impossible to create: no feature selected")
    
        
    def timeseries_dag(self,
                       min_width = 1,
                       max_width = 5,
                       min_score = 0,
                       max_score = 1,
                       node_size = 8,
                       font_size = 12,
                       node_color = 'orange',
                       edge_color = 'grey'):
        """
        Saves timeseries dag plot if resfolder has been set otherwise it shows the figure
        
        Args:
            min_width (int, optional): minimum linewidth. Defaults to 1.
            max_width (int, optional): maximum linewidth. Defaults to 5.
            min_score (int, optional): minimum score range. Defaults to 0.
            max_score (int, optional): maximum score range. Defaults to 1.
            node_size (int, optional): node size. Defaults to 8.
            node_color (str, optional): node color. Defaults to 'orange'.
            edge_color (str, optional): edge color. Defaults to 'grey'.
            font_size (int, optional): font size. Defaults to 12.
        """
        
        if self.result:
            self.validator.build_ts_dag(min_width,
                                        max_width,
                                        min_score,
                                        max_score,
                                        node_size,
                                        node_color,
                                        edge_color,
                                        font_size)
        else:
            CP.warning("Timeseries dag impossible to create: no feature selected")


    def get_selected_features(self):
        """
        Defines the list of selected variables for d

        Returns:
            list(str): list of selected variable names
        """
        f_list = list()
        for t in self.filter_dependencies:
            sources_t = self.__get_dependencies_for_target(t)
            if self.neglect_only_autodep and self.__is_only_autodep(sources_t, t):
                sources_t.remove(t)
            if sources_t: sources_t.append(t)
            f_list = list(set(f_list + sources_t))
        res = [f for f in self.data.features if f in f_list]

        return res
    
    
    def show_dependencies(self):
        """
        Saves dependencies graph if resfolder is set otherwise it shows the figure
        """
        # FIXME: LAG not considered
        dependencies_matrix = self.__get_dependencies_matrix()

        fig, ax = plt.subplots()
        im = ax.imshow(dependencies_matrix, cmap=plt.cm.Greens, interpolation='nearest', vmin=0, vmax=1, origin='lower')
        fig.colorbar(im, orientation='vertical', label="score")

        plt.xlabel("Sources")
        plt.ylabel("Targets")
        plt.xticks(ticks = range(0, self.data.orig_N), labels = self.data.orig_pretty_features, fontsize = 8)
        plt.yticks(ticks = range(0, self.data.orig_N), labels = self.data.orig_pretty_features, fontsize = 8)
        plt.title("Dependencies")

        if self.dependency_path is not None:
            plt.savefig(self.dependency_path, dpi = 300)
        else:
            plt.show()


    def print_dependencies(self):
        """
        Print dependencies found by the selector
        """
        for t in self.o_filter_dependecies:
            print()
            print()
            print(DASH)
            print("Target", t)
            print(DASH)
            print('{:<10s}{:>15s}{:>15s}{:>15s}'.format('SOURCE', 'SCORE', 'PVAL', 'LAG'))
            print(DASH)
            for s in self.o_filter_dependecies[t]:
                print('{:<10s}{:>15.3f}{:>15.3f}{:>15d}'.format(s[SOURCE], s[SCORE], s[PVAL], s[LAG]))      


    def load_result(self, res_path):
        with open(res_path, 'rb') as f:
            self.validator.result = pickle.load(f)


    def __shrink_dependencies(self):
        """
        Shrinks dependencies based on the selected features
        """
        difference_set = self.filter_dependencies.keys() - self.data.features
        for d in difference_set: 
            del self.filter_dependencies[d]
            if self.validator.dependencies is not None: del self.validator.dependencies['$' + d + '$']
 

    def __get_dependencies_for_target(self, t):
        """
        Returns list of sources for a specified target

        Args:
            t (str): target variable name

        Returns:
            list(str): list of sources for target t
        """
        return [s[SOURCE] for s in self.filter_dependencies[t]]
    
    
    def __is_only_autodep(self, sources, t):
        """
        Returns list of sources for a specified target

        Args:
            sources (list(str)): list of sources for the selected target
            t (str): target variable name

        Returns:
            (bool): True if sources list contains only the target. False otherwise
        """
        if len(sources) == 1 and sources[0] == t: return True
        return False


    def __get_dependencies_matrix(self):
        """
        Returns a matrix composed by scores for each target

        Returns:
            (np.array): score matrix
        """
        dep_mat = list()
        for t in self.o_filter_dependecies:
            dep_vet = [0] * self.data.orig_N
            for s in self.o_filter_dependecies[t]:
                dep_vet[self.data.orig_features.index(s[SOURCE])] = s[SCORE]
            dep_mat.append(dep_vet)

        dep_mat = np.array(dep_mat)
        inf_mask = np.isinf(dep_mat)
        neginf_mask = np.isneginf(dep_mat)
        max_dep_mat = np.max(dep_mat[(dep_mat != -np.inf) & (dep_mat != np.inf)])
        min_dep_mat = np.min(dep_mat[(dep_mat != -np.inf) & (dep_mat != np.inf)])

        dep_mat[inf_mask] = max_dep_mat
        dep_mat[neginf_mask] = min_dep_mat
        dep_mat = (dep_mat - min_dep_mat) / (max_dep_mat - min_dep_mat)
        return dep_mat


    def __get_link_assumptions(self):
        """
        Return selected links found by the selector
        in this form: {0: {(0,-1) : "-?>", (2,-1) : "-?>"}}

        Returns:
            (dict): selected links
        """
        sel_links = {self.data.features.index(f):dict() for f in self.data.features}
        for t in self.filter_dependencies:
            
            # add links
            for s in self.filter_dependencies[t]:
                sel_links[self.data.features.index(t)][(self.data.features.index(s[SOURCE]), -s[LAG])] = '-?>'

        return sel_links
    
    
    def __apply_validator_result(self, causal_model):
        """
        Exclude dependencies based on validator result
        It does not overwrite the filter_dependencies' inference/p-values matrix with the ones found by the validator
        """
        list_diffs = list()
        tmp_dependencies = copy.deepcopy(self.filter_dependencies)
        for t in tmp_dependencies:
            for s in tmp_dependencies[t]:
                if (self.data.features.index(s[SOURCE]), -s[LAG]) not in causal_model[self.data.features.index(t)]:
                    list_diffs.append((s[SOURCE], str(s[LAG]), t))
                    self.filter_dependencies[t].remove(s)
        if list_diffs:
            CP.debug(DASH)
            CP.debug("Difference(s)")
            CP.debug(DASH)
            for diff in list_diffs:
                CP.debug("Removing (" + diff[0] + " -" + diff[1] +") --> (" + diff[2] + ")")
    



from enum import Enum

SOURCE = 'source'
SCORE = 'score'
PVAL = 'pval'
LAG = 'lag'

DASH = '-' * 55
RESFOLDER_DEFAULT = "%Y-%m-%d_%H-%M-%S"
SEP = "/"
RES_FILENAME = "res.pkl"
DAG_FILENAME = "dag.png"
TSDAG_FILENAME = "ts_dag.pdf"
DEP_FILENAME = "dependency.pdf"
LOG_FILENAME = "log.txt"
RESULTS_FILENAME = "results"

class LabelType(Enum):
    Lag = "Lag"
    Score = "Score"
    NoLabels = "NoLabels"


import math
from scipy.stats import entropy


class MovingWindow:
    def __init__(self, window):
        self.window = window
        self.T, self.dim = window.shape
        self.entropy = None
        self.opt_size = None
        self.opt_samples_index = None

    def get_pdf(self):
        """
        Compute the probability distribution function from an array of data

        Returns:
            list: probability distribution function
        """

        counts = {}

        for i in range(0, self.T):
            t = tuple(self.window[i, :])
            if t in counts:
                counts[t] += 1
            else:
                counts[t] = 1

        pdf = {k: v / self.T for k, v in counts.items()}

        return list(pdf.values())


    def get_entropy(self):
        """
        Compute the entropy based on probability distribution function
        """
        self.entropy = entropy(self.get_pdf(), base = 2)


    def samples_selector(self, step):
        """
        Select sample to be taken from a moving window


        Args:
            step (int): subsampling frequency

        Returns:
            list[int]: list of indexes corresponding to the sample to be taken
        """
        return [i for i in range(0, self.T, step)]


    def optimal_sampling(self, thres):
        """
        Find the optimal number of sample for a particular moving window


        Args:
            thres (float): stopping criteria threshold
        """
        converged = False
        _old_step = 0
        _sub_index = list(range(0, self.T))
        _old_sub_index = list(range(0, self.T))
        _max_n = math.floor(self.T / 2)

        for n in range(_max_n, 1, -1):
            # resampling window with n samples and build another Moving Window
            step = int(self.T / n)
            if step == _old_step:
                continue
            _old_step = step
            _old_sub_index = _sub_index
            _sub_index = self.samples_selector(step)
            _sub_w = MovingWindow(self.window[_sub_index])

            # compute entropy on the sub window
            _sub_w.get_entropy()

            # stopping criteria
            if self.entropy != 0:
                if abs(_sub_w.entropy - self.entropy) / self.entropy >= thres:
                    converged = True
                    break
        self.opt_size = len(_old_sub_index) if converged else len(_sub_index)
        self.opt_samples_index = _old_sub_index if converged else _sub_index



from fpcmci.version import VERSION

"""
FPCMCI.

Filtered - PCMCI causal discovery algorithm.
Extension of the PCMCI causal discovery algorithm augmented with a feature selection method.
"""

__version__ = VERSION
__author__ = 'Luca Castri'


import numpy as np
import pandas as pd
from matplotlib import gridspec
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler
from fpcmci.preprocessing.Subsampler import Subsampler
from fpcmci.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod


class Data():
    """
    Data class manages the preprocess of the data before the causal analysis
    """
    def __init__(self, data, vars = None, fill_nan = True, stand = False, subsampling : SubsamplingMethod = None, show_subsampling = False):
        """
        Data class constructor

        Args:
            data (str/DataFrame/np.array): it can be a string specifing the path of a csv file to load/pandas.DataFrame/numpy.array
            vars (list(str), optional): List containing variable names. If unset then, 
                if data = (str/DataFrame) vars = data columns name elif data = np.array vars = [X_0 .. X_N]
                Defaults to None.
            fill_nan (bool, optional): Fill NaNs bit. Defaults to True.
            stand (bool, optional): Standardization bit. Defaults to False.
            subsampling (SubsamplingMethod, optional): Subsampling method. If None not active. Defaults to None.
            show_subsampling (bool, optional): If True shows subsampling result. Defaults to False.

        Raises:
            TypeError: if data is not str - DataFrame - ndarray
        """
        # Data handling
        if type(data) == np.ndarray:
            self.d = pd.DataFrame(data)
            if vars is None: self.d.columns = list(['X_' + str(f) for f in range(len(self.d.columns))])
        elif type(data) == pd.DataFrame:
            self.d = data
        elif type(data) == str:
            self.d = pd.read_csv(data)
        else:
            raise TypeError("data field not in the correct type\ndata must be one of the following type:\n- numpy.ndarray\n- pandas.DataFrame\n- .csv path")
            
        
        # Columns name handling
        if vars is not None:
            self.d.columns = list(vars)
                
        
        self.orig_features = self.features
        self.orig_pretty_features = self.pretty_features
        self.orig_N = self.N
        self.orig_T = len(self.d)

        # Filling NaNs
        if fill_nan:
            if self.d.isnull().values.any():
                self.d.fillna(inplace=True, method="ffill")
                self.d.fillna(inplace=True, method="bfill")

        # Subsampling data
        if subsampling is not None:
            subsampler = Subsampler(self.d, ss_method = subsampling)
            self.d = pd.DataFrame(subsampler.subsample(), columns = self.features)
            if show_subsampling: subsampler.plot_subsampled_data()

        # Standardize data
        if stand:
            scaler = StandardScaler()
            scaler = scaler.fit(self.d)
            self.d = pd.DataFrame(scaler.transform(self.d), columns = self.features)
        
        
    @property  
    def features(self):
        """
        Returns list of features

        Returns:
            list(str): list of feature names
        """
        return list(self.d.columns)

    @property
    def pretty_features(self):
        """
        Returns list of features with LATEX symbols
                
        Returns:
            list(str): list of feature names
        """
        return [r'$' + str(v) + '$' for v in self.d.columns]
    
    @property
    def N(self):
        """
        Number of features
        
        Returns:
            (int): number of features
        """
        return len(self.d.columns)

    @property
    def T(self):
        """
        Dataframe length
        
        Returns:
            (int): dataframe length
        """
        return len(self.d)
                       
            
    def shrink(self, selected_features):
        """
        Shrinks dataframe d and dependencies based on the selected features

        Args:
            selected_features (list(str)): features selected by the selector
        """
        self.d = self.d[selected_features]
        
                    
    def plot_timeseries(self):
        """
        Plots timeseries data
        """
        # Create grid
        gs = gridspec.GridSpec(self.N, 1)

        # Time vector
        T = list(range(self.T))

        plt.figure()
        for i in range(0, self.d.shape[1]):
            ax = plt.subplot(gs[i, 0])
            plt.plot(T, self.d.values[:, i], color = 'tab:red')
            plt.ylabel(str(self.pretty_features[i]))

        plt.show()


from scipy.fft import rfft, rfftfreq
from fpcmci.preprocessing.subsampling_methods.EntropyBasedMethod import EntropyBasedMethod
from fpcmci.preprocessing.subsampling_methods.SubsamplingMethod import SSMode, SubsamplingMethod
import numpy as np
import pylab as pl
from math import ceil
import scipy.signal


class WSFFTStatic(SubsamplingMethod, EntropyBasedMethod):
    """
    Subsampling method with static window size based on Fourier analysis
    """
    def __init__(self, sampling_time, entropy_threshold):
        """
        WSFFTStatic class constructor

        Args:
            sampling_time (float): timeseries sampling time
            entropy_threshold (float): entropy threshold
        """
        SubsamplingMethod.__init__(self, SSMode.WSFFTStatic)
        EntropyBasedMethod.__init__(self, entropy_threshold)
        self.sampling_time = sampling_time


    def __fourier_window(self):
        """
        Compute window size based on Fourier analysis performed on dataframe

        Returns:
            (int): window size
        """
        N, dim = self.df.shape
        xf = rfftfreq(N, self.sampling_time)
        w_array = list()
        for i in range(0, dim):
            yf = np.abs(rfft(self.df.values[:, i]))

            peak_indices, _ = scipy.signal.find_peaks(yf)
            highest_peak_index = peak_indices[np.argmax(yf[peak_indices])]
            w_array.append(ceil(1 / (2 * xf[highest_peak_index]) / self.sampling_time))
            # fig, ax = pl.subplots()
            # ax.plot(xf, yf)
            # ax.plot(xf[highest_peak_index], np.abs(yf[highest_peak_index]), "x")
            # pl.show()
        return min(w_array)


    def dataset_segmentation(self):
        """
        Segments dataset with a fixed window size
        """
        seg_res = [i for i in range(0, len(self.df.values), self.ws)]
        self.segments = [(i, i + self.ws) for i in range(0, len(self.df.values) - self.ws, self.ws)]
        if not seg_res.__contains__(len(self.df.values)):
            self.segments.append((seg_res[-1], len(self.df.values)))
            seg_res.append(len(self.df.values))

    
    def run(self):
        """
        Run subsampler

        Returns:
            (list[int]): indexes of the remaining samples
        """
        # define window size
        self.ws = self.__fourier_window()

        # build list of segment
        self.dataset_segmentation()

        # compute entropy moving window
        self.moving_window_analysis()

        # extracting subsampling procedure results
        idxs = self.extract_indexes()

        return idxs


import ruptures as rpt
from fpcmci.preprocessing.subsampling_methods.EntropyBasedMethod import EntropyBasedMethod
from fpcmci.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod, SSMode


class WSDynamic(SubsamplingMethod, EntropyBasedMethod):
    """
    Subsampling method with dynamic window size based on entropy analysis
    """
    def __init__(self, window_min_size, entropy_threshold):
        """
        WSDynamic class constructor

        Args:
            window_min_size (int): minimun window size
            entropy_threshold (float): entropy threshold

        Raises:
            ValueError: if window_min_size == None
        """
        SubsamplingMethod.__init__(self, SSMode.WSDynamic)
        EntropyBasedMethod.__init__(self, entropy_threshold)
        if window_min_size is None:
            raise ValueError("window_type = DYNAMIC but window_min_size not specified")
        self.wms = window_min_size
        self.ws = None

    def dataset_segmentation(self):
        """
        Segments dataset based on breakpoint analysis and a min window size
        """
        de = self.create_rounded_copy()
        algo = rpt.Pelt(model = "l2", min_size = self.wms).fit(de)
        seg_res = algo.predict(pen = 10)
        self.segments = [(seg_res[i - 1], seg_res[i]) for i in range(1, len(seg_res))]
        self.segments.insert(0, (0, seg_res[0]))


    def run(self):
        """
        Run subsampler

        Returns:
            (list[int]): indexes of the remaining samples
        """
        # build list of segment
        self.dataset_segmentation()

        # compute entropy moving window
        self.moving_window_analysis()

        # extracting subsampling procedure results
        idxs = self.extract_indexes()

        return idxs


from fpcmci.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod, SSMode


class Static(SubsamplingMethod):
    """
    Subsamples data by taking one sample each step-samples
    """
    def __init__(self, step):
        """
        Static class constructor
        
        Args:
            step (int): integer subsampling step

        Raises:
            ValueError: if step == None
        """
        super().__init__(SSMode.Static)
        if step is None:
            raise ValueError("step not specified")
        self.step = step

    def run(self):
        return range(0, len(self.df.values), self.step)


from fpcmci.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod
import pandas as pd
from matplotlib import gridspec
import pylab as pl
import numpy as np

class Subsampler():
    """
    Subsampler class. 
    
    It subsamples the data by using a subsampling method chosen among:
        - Static - subsamples data by taking one sample each step-samples
        - WSDynamic - entropy based method with dynamic window size computed by breakpoint analysis
        - WSFFTStatic - entropy based method with fixed window size computed by FFT analysis
        - WSStatic - entropy base method with predefined window size
    """

    def __init__(self, 
                 df: pd.DataFrame, 
                 ss_method: SubsamplingMethod):
        """
        Subsampler class constructor

        Args:
            df (pd.DataFrame): dataframe to subsample
            ss_method (SubsamplingMethod): subsampling method
        """
        self.df = df
        self.ss_method = ss_method
        self.ss_method.initialise(df)


    def subsample(self):
        """
        Runs the subsampling algorithm and returns the subsapled ndarray

        Returns:
            (ndarray): Subsampled dataframe value
        """
        self.result = self.ss_method.run()
        return self.df.values[self.result, :]


    def plot_subsampled_data(self, dpi = 100, show = True):
        """
        Plot dataframe sub-sampled data

        Args:
            dpi (int, optional): image dpi. Defaults to 100.
            show (bool, optional): if True it shows the figure and block the process. Defaults to True.
        """
        n_plot = self.df.shape[1]

        # Create grid
        gs = gridspec.GridSpec(n_plot, 1)

        # Time vector
        T = list(range(0, self.df.shape[0]))

        pl.figure(dpi = dpi)
        for i in range(0, n_plot):
            ax = pl.subplot(gs[i, 0])
            pl.plot(T, self.df.values[:, i], color = 'tab:red')
            pl.scatter(np.array(T)[self.result],
                       self.df.values[self.result, i],
                       s = 80,
                       facecolors = 'none',
                       edgecolors = 'b')
            pl.gca().set(ylabel = r'$' + str(self.df.columns.values[i]) + '$')
        if show:
            pl.show()


import copy
import pickle
from tigramite.pcmci import PCMCI as VAL
from tigramite.independence_tests.independence_tests_base import CondIndTest
import tigramite.data_processing as pp
import numpy as np
from fpcmci.CPrinter import CPLevel, CP
from fpcmci.preprocessing.data import Data
import fpcmci.basics.utils as utils
from fpcmci.causal_graph import *
from fpcmci.basics.constants import *


class PCMCI():
    """
    PCMCI class.

    PCMCI works with FSelector in order to find the causal 
    model starting from a prefixed set of variables and links.
    """
    def __init__(self, data: Data, alpha, min_lag, max_lag, val_condtest: CondIndTest, resfolder, verbosity: CPLevel):
        """
        PCMCI class constructor

        Args:
            data (Data): data to analyse
            alpha (float): significance level
            min_lag (int): minimum time lag
            max_lag (int): maximum time lag
            val_condtest (CondIndTest): validation method
            resfolder (str): result folder. If None then the results are not saved.
            verbosity (CPLevel): verbosity level
        """
        self.data = data
        self.alpha = alpha
        self.min_lag = min_lag
        self.max_lag = max_lag
        self.result = None
        self.dependencies = None
        self.val_method = None
        self.val_condtest = val_condtest
        self.verbosity = verbosity.value

        self.respath = None
        self.dag_path = None
        self.ts_dag_path = None
        if resfolder is not None:
            self.respath, self.dag_path, self.ts_dag_path = utils.get_validatorpaths(resfolder)  
        

    def run(self, link_assumptions = None):
        """
        Run causal discovery algorithm

        Returns:
            (dict): estimated causal model
        """
        CP.info('\n')
        CP.info(DASH)
        CP.info("Running Causal Discovery Algorithm")

        # build tigramite dataset
        vector = np.vectorize(float)
        data = vector(self.data.d)
        dataframe = pp.DataFrame(data = data,
                                 var_names = self.data.pretty_features)
        
        # init and run pcmci
        self.val_method = VAL(dataframe = dataframe,
                                cond_ind_test = self.val_condtest,
                                verbosity = self.verbosity)

        self.result = self.val_method.run_pcmci(link_assumptions = link_assumptions,
                                                tau_max = self.max_lag,
                                                tau_min = self.min_lag)
        
        self.result['var_names'] = self.data.pretty_features
        # apply significance level
        self.result['graph'] = self.__apply_alpha()
        self.dependencies = self.__PCMCIres_converter()
        return self.__return_parents_dict()
    
    
    def build_ts_dag(self,
                     min_width,
                     max_width,
                     min_score,
                     max_score,
                     node_size,
                     node_color,
                     edge_color,
                     font_size):
        """
        Saves timeseries dag plot if resfolder is set otherwise it shows the figure
        
        Args:
            min_width (int): minimum linewidt
            max_width (int): maximum linewidth
            min_score (int): minimum score range
            max_score (int): maximum score range
            node_size (int): node size
            node_color (str): node color
            edge_color (str): edge color 
            font_size (int): font size
        """
        
        # # convert to dictionary
        # res = self.__PCMCIres_converter()
        
        # # filter only dependencies
        # tmp_res = {k: res[k] for k in self.data.pretty_features}
        # res = tmp_res
        
        ts_dag(self.dependencies, 
               tau = self.max_lag,
               min_width = min_width,
               max_width = max_width,
               min_score = min_score,
               max_score = max_score,
               node_size = node_size,
               node_color = node_color,
               edge_color = edge_color,
               font_size = font_size,
               save_name = self.ts_dag_path)


    def build_dag(self,
                  node_layout,
                  min_width,
                  max_width,
                  min_score,
                  max_score,
                  node_size,
                  node_color,
                  edge_color,
                  font_size,
                  label_type):
        """
        Saves dag plot if resfolder is set otherwise it shows the figure
        
        Args:
            node_layout (str): node_layout
            min_width (int): minimum linewidth
            max_width (int): maximum linewidth
            min_score (int): minimum score range
            max_score (int): maximum score range
            node_size (int): node size
            node_color (str): node color
            edge_color (str): edge color
            font_size (int): font size
            label_type (LabelType, optional): enum to set whether to show the lag time (LabelType.Lag) or the strength (LabelType.Score) of the dependencies on each link/node or not showing the labels (LabelType.NoLabels). Default LabelType.Lag.

        """               
        
        # # convert to dictionary
        # res = self.__PCMCIres_converter()
        
        # # filter only dependencies
        # tmp_res = {k: res[k] for k in self.data.pretty_features}
        # res = tmp_res
        
        dag(self.dependencies,
            node_layout = node_layout,
            min_width = min_width,
            max_width = max_width,
            min_score = min_score,
            max_score = max_score,
            node_size = node_size,
            font_size = font_size,
            node_color = node_color,
            edge_color = edge_color,
            label_type = label_type,
            save_name = self.dag_path)
        
        
    def save_result(self):
        """
        Save causal discovery results as pickle file if resfolder is set
        """
        if self.respath is not None:
            res = dict()
            res['dependencies'] = copy.deepcopy(self.dependencies)
            # res = copy.deepcopy(self.result)
            res['alpha'] = self.alpha
            res['var_names'] = self.data.pretty_features
            res['dag_path'] = self.dag_path
            res['ts_dag_path'] = self.ts_dag_path
            with open(self.respath, 'wb') as resfile:
                pickle.dump(res, resfile)
        
        
    def __return_parents_dict(self):
        """
        Returns dictionary of parents sorted by val_matrix filtered by alpha

        Returns:
            (dict): Dictionary of form {0:[(0, -1), (3, -2), ...], 1:[], ...} containing estimated parents.
        """
        graph = self.result['graph']
        val_matrix = self.result['val_matrix']
        p_matrix = self.result['p_matrix']
        
        # Initialize the return value
        parents_dict = dict()
        for j in range(self.data.N):
            # Get the good links
            good_links = np.argwhere(graph[:, j, 1:] == "-->")
            # Build a dictionary from these links to their values
            links = {(i, -tau - 1): np.abs(val_matrix[i, j, abs(tau) + 1]) 
                     for i, tau in good_links if p_matrix[i, j, abs(tau) + 1] <= self.alpha}
            # Sort by value
            parents_dict[j] = sorted(links, key=links.get, reverse=True)
        
        return parents_dict
    
    
    def __PCMCIres_converter(self):
        """
        Re-elaborates the PCMCI result in a new dictionary

        Returns:
            (dict): pcmci result re-elaborated
        """
        res_dict = {f:list() for f in self.result['var_names']}
        N, lags = self.result['graph'][0].shape
        for s in range(len(self.result['graph'])):
            for t in range(N):
                for lag in range(lags):
                    if self.result['graph'][s][t,lag] == '-->':
                        res_dict[self.result['var_names'][t]].append({SOURCE : self.result['var_names'][s],
                                                                      SCORE : self.result['val_matrix'][s][t,lag],
                                                                      PVAL : self.result['p_matrix'][s][t,lag],
                                                                      LAG : lag})
        return res_dict
    

    def __apply_alpha(self):
        """
        Applies alpha threshold to the pcmci result

        Returns:
            (ndarray): graph filtered by alpha 
        """
        mask = np.ones(self.result['p_matrix'].shape, dtype='bool')
        
        # Set all p-values of absent links to 1.
        self.result['p_matrix'][mask==False] == 1.
        
        # Threshold p_matrix to get graph
        graph_bool = self.result['p_matrix'] <= self.alpha
        
        # Convert to string graph representation
        graph = self.__convert_to_string_graph(graph_bool)
        
        return graph
    
    
    def __convert_to_string_graph(self, graph_bool):
        """
        Converts the 0,1-based graph returned by PCMCI to a string array
        with links '-->'

        Args:
            graph_bool (array): 0,1-based graph array output by PCMCI

        Returns:
            (array): graph as string array with links '-->'.
        """

        graph = np.zeros(graph_bool.shape, dtype='<U3')
        graph[:] = ""
        # Lagged links
        graph[:,:,1:][graph_bool[:,:,1:]==1] = "-->"
        # Unoriented contemporaneous links
        graph[:,:,0][np.logical_and(graph_bool[:,:,0]==1, 
                                    graph_bool[:,:,0].T==1)] = "o-o"
        # Conflicting contemporaneous links
        graph[:,:,0][np.logical_and(graph_bool[:,:,0]==2, 
                                    graph_bool[:,:,0].T==2)] = "x-x"
        # Directed contemporaneous links
        for (i,j) in zip(*np.where(
            np.logical_and(graph_bool[:,:,0]==1, graph_bool[:,:,0].T==0))):
            graph[i,j,0] = "-->"
            graph[j,i,0] = "<--"
        return graph


from setuptools import setup
from fpcmci.version import VERSION

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


INSTALL_REQUIRES = ["tigramite>=5.1.0.3",
                    "pandas>=1.5.2",
                    "netgraph>=4.10.2",
                    "networkx>=2.8.6",
                    "ruptures>=1.1.7",
                    "scikit_learn==1.1.3",
                    "torch>=1.11.0", 
                    "gpytorch>=1.4",       
                    "dcor>=0.5.3",
                    "h5py>=3.7.0"   
                    ]

setup(
    name = 'fpcmci',
    version = VERSION,    
    description = 'A causal discovery Python package',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/lcastri/fpcmci',
    author = 'Luca Castri',
    author_email = 'lucacastri94@gmail.com',
    packages = ['fpcmci', "fpcmci.preprocessing", "fpcmci.preprocessing.subsampling_methods", "fpcmci.basics", "fpcmci.selection_methods"],
    python_requires='>=3',
    install_requires = INSTALL_REQUIRES,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',  
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
)



import sys
from fpcmci.basics.utils import cls


class Logger(object):
    def __init__(self, path):
        self.terminal = sys.stdout
        self.log = open(path, "w")
        cls()

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass


from enum import Enum
import numpy as np
from fpcmci.selection_methods.SelectionMethod import SelectionMethod, CTest, _suppress_stdout
from idtxl.multivariate_te import MultivariateTE
from idtxl.bivariate_mi import BivariateMI
from idtxl.data import Data
from fpcmci.CPrinter import CP


class TEestimator(Enum):
    Gaussian = 'JidtGaussianCMI'
    Kraskov = 'JidtKraskovCMI'


class TE(SelectionMethod):
    """
    Feature selection method based on Trasfer Entropy analysis
    """
    def __init__(self, estimator: TEestimator):
        """
        TE class contructor

        Args:
            estimator (TEestimator): Gaussian/Kraskov
        """
        super().__init__(CTest.TE)
        self.estimator = estimator


    def compute_dependencies(self):
        """
        compute list of dependencies for each target by transfer entropy analysis

        Returns:
            (dict): dictonary(TARGET: list SOURCES)
        """
        multi_network_analysis = MultivariateTE()
        bi_network_analysis = BivariateMI()
        settings = {'cmi_estimator': self.estimator.value,
                    'max_lag_sources': self.max_lag,
                    'min_lag_sources': self.min_lag,
                    'max_lag_target': self.max_lag,
                    'min_lag_target': self.min_lag,
                    'alpha_max_stats': self.alpha,
                    'alpha_min_stats': self.alpha,
                    'alpha_omnibus': self.alpha,
                    'alpha_max_seq': self.alpha,
                    'verbose': False}
        
        CP.info("\n##")
        CP.info("## " + self.name + " analysis")
        CP.info("##")
        for target in self.data.features:
            CP.info("\n## Target variable: " + target)
            with _suppress_stdout():
                t = self.data.features.index(target)
                
                # Check auto-dependency
                tmp_d = np.c_[self.data.d.values[:, t], self.data.d.values[:, t]]
                data = Data(tmp_d, dim_order='sp') # sp = samples(row) x processes(col)
                res_auto = bi_network_analysis.analyse_single_target(settings = settings, data = data, target = 0, sources = 1)
                
                # Check cross-dependencies
                data = Data(self.data.d.values, dim_order='sp') # sp = samples(row) x processes(col)
                res_cross = multi_network_analysis.analyse_single_target(settings = settings, data = data, target = t)
            
            # Auto-dependency handling
            auto_lag = [s[1] for s in res_auto._single_target[0]['selected_vars_sources']]
            auto_score = res_auto._single_target[0]['selected_sources_mi']
            auto_pval = res_auto._single_target[0]['selected_sources_pval']
            if auto_score is not None:
                for score, pval, lag in zip(auto_score, auto_pval, auto_lag):
                    self._add_dependecies(self.data.features[t], self.data.features[t], score, pval, lag)
            
            # Cross-dependencies handling    
            sel_sources = [s[0] for s in res_cross._single_target[t]['selected_vars_sources']]
            if sel_sources:
                sel_sources_lag = [s[1] for s in res_cross._single_target[t]['selected_vars_sources']]
                sel_sources_score = res_cross._single_target[t]['selected_sources_te']
                sel_sources_pval = res_cross._single_target[t]['selected_sources_pval']
                for s, score, pval, lag in zip(sel_sources, sel_sources_score, sel_sources_pval, sel_sources_lag):
                    self._add_dependecies(self.data.features[t], self.data.features[s], score, pval, lag)
            
            if auto_score is None and not sel_sources:
                CP.info("no sources selected")

        return self.result


from fpcmci.preprocessing.subsampling_methods.EntropyBasedMethod import EntropyBasedMethod
from fpcmci.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod, SSMode


class WSStatic(SubsamplingMethod, EntropyBasedMethod):
    """
    Entropy based subsampling method with static window size
    """
    def __init__(self, window_size, entropy_threshold):
        """
        WSStatic class constructor
        
        Args:
            window_size (int): minimun window size
            entropy_threshold (float): entropy threshold

        Raises:
            ValueError: if window_size == None
        """
        
        SubsamplingMethod.__init__(self, SSMode.WSDynamic)
        EntropyBasedMethod.__init__(self, entropy_threshold)
        if window_size is None:
            raise ValueError("window_type = STATIC but window_size not specified")
        self.ws = window_size


    def dataset_segmentation(self):
        """
        Segments dataset with a fixed window size
        """
        seg_res = [i for i in range(0, len(self.df.values), self.ws)]
        self.segments = [(i, i + self.ws) for i in range(0, len(self.df.values) - self.ws, self.ws)]
        if not seg_res.__contains__(len(self.df.values)):
            self.segments.append((seg_res[-1], len(self.df.values)))
            seg_res.append(len(self.df.values))


    def run(self):
        """
        Run subsampler

        Returns:
            (list[int]): indexes of the remaining samples
        """
        # build list of segment
        self.dataset_segmentation()

        # compute entropy moving window
        self.moving_window_analysis()

        # extracting subsampling procedure results
        idxs = self.extract_indexes()

        return idxs